
package com.sap.dbtechdev.wrapper;

import java.lang.reflect.*;
import java.io.PrintWriter;

/**
 *
 */
class InterfaceWrapper
{
    protected WrapperOptions options;
    protected String interfacePackage;
    protected String [] interfaceClasses;
    /**
     * InterfaceWrapper constructor
     */
    InterfaceWrapper (
        WrapperOptions optionsP,
        String interfacePackageP,
        String [] interfaceClassesP)
    {
        this.options = optionsP;
        this.interfacePackage = interfacePackageP;
        this.interfaceClasses = interfaceClassesP;
    }
    /**
     *
     */
    public void genWrapper ()
    throws java.io.IOException
    {
         for (int i = 0; i < this.interfaceClasses.length; ++i) {
             String fullName;
             String baseName = this.interfaceClasses [i];

             if (this.interfacePackage != null) {
                 fullName = this.interfacePackage + "."
                     + baseName;
             }
             else {
                 fullName = baseName;
             }
             try {
                 this.genClassWrapper (fullName, baseName);
             }
             catch (ClassNotFoundException exc) {
                 System.out.println ("Not found: " + fullName);
             }
         }
    }
    /**
     *
     */
    protected void genClassWrapper (
        String fullName,
        String baseName)
    throws ClassNotFoundException, java.io.IOException
    {
        Class clazz = Class.forName (fullName);
        PrintWriter outstream = this.options.getSourceStream (baseName);
        this.genIntro (outstream);
        this.genClassDeclaration (fullName, baseName, outstream);
        this.genConstructor (fullName, baseName, outstream);
        Method [] methods = clazz.getMethods ();
        for (int i = 0; i < methods.length; ++i) {
            this.genMethodWrapper (methods [i], outstream);
        }
        this.genExit (outstream);
        outstream.close ();
    }
    /**
     *
     */
    protected void genMethodWrapper (
        Method method,
        PrintWriter outstream)
    {
        Class returnType = method.getReturnType ();
        String name = method.getName ();
        Class [] parameterTypes = method.getParameterTypes ();
        Class [] exceptionTypes = method.getExceptionTypes ();
        /*
         * genDeclaration
         */
        String delimiter = "";
        outstream.print ("    public " + this.typeString (returnType)
            + " " + name + " (");
        for (int i = 0; i < parameterTypes.length; ++i) {
            outstream.println (delimiter);
            outstream.print ("        " + this.typeString (parameterTypes [i])
                + " arg" + i);
            delimiter = ",";
        }
        outstream.println (")");
        /*
         * gen exceptions
         */
        if (exceptionTypes.length > 0) {
            outstream.print ("    throws "
                + this.typeString (exceptionTypes [0]));
            for (int i = 1; i < exceptionTypes.length; ++i) {
                outstream.print (",        "
                    + this.typeString (exceptionTypes [i]));
            }
            outstream.println ();
        }
        outstream.println ("    {");
        /*
         * declare result var
         */
        if (returnType != Void.TYPE) {
            outstream.println ("        " + this.typeString (returnType) + " _result;");
        }
        /*
         * gen parameter entry
         */
        this.genParameterEntry (method, outstream);
        outstream.println ("        try {");
        /*
         * gen inner call
         */
        outstream.print ("            ");
        if (returnType != Void.TYPE) {
            outstream.print ("_result = ");
        }
        outstream.print ("this._inner." + name + " (");
        delimiter = "arg";
        for (int i = 0; i < parameterTypes.length; ++i) {
            outstream.print (delimiter + i);
            delimiter = ", arg";
        }
        outstream.println (");");
        /*
         * optional: wrap result
         */
        if ((returnType != Void.TYPE) && (this.generatesClass (returnType))) {
            String clazzName = returnType.getName ();
            int lastPoint = clazzName.lastIndexOf ('.');
            String shortName = clazzName.substring (lastPoint + 1);
            outstream.print ("            _result = new "
                + this.options.getTargetPackage () + "."
                + shortName
                + " (_result");
            HelperVar[] helperVars = this.options.getHelperVars ();
            for (int i = 0; i < helperVars.length; ++i) {
                outstream.print (", this." + helperVars [i].getVarName ());
            }
            outstream.println (");");
        }
        outstream.println ("        }");
        outstream.println ("        catch (RuntimeException _exc) {");
        this.genExceptionHandling (method, "_exc", outstream);
        outstream.println ("            throw _exc;");
        outstream.println ("        }");
        for (int i = 0; i < exceptionTypes.length; ++i) {
            outstream.println ("        catch (" + exceptionTypes [i].getName () + " _exc" + i + ") {");
            this.genExceptionHandling (method, "_exc" + i, outstream);
            outstream.println ("            throw _exc" + i + ";");
            outstream.println ("        }");
        }
        /*
         * return result
         */
        if (returnType != Void.TYPE) {
            this.genReturn (method, "_result", outstream);
            outstream.println ("        return _result;");
        }
        outstream.println ("    }");
        outstream.println ();
    }

    /**
     *
     */
    protected void genIntro (
        PrintWriter outstream)
    {
        String targetPackage = this.options.getTargetPackage ();
        if (targetPackage != null) {
            outstream.print ("\npackage " + targetPackage + ";\n\n");
        }
        this.genImports (outstream);
    }
    /**
     *
     */
    protected void genExit (
        PrintWriter outstream)
    {
        outstream.println ("\n}");
    }
    /**
     *
     */
    protected void genClassDeclaration (
        String fullName,
        String baseName,
        PrintWriter outstream)
    {
        outstream.println ("public class " + baseName + " implements " + fullName);
        outstream.println ("{");
        this.genInstanceVar ("_inner", fullName, outstream);
        HelperVar[] helperVars = this.options.getHelperVars ();
        for (int i = 0; i < helperVars.length; ++i) {
            this.genInstanceVar (helperVars [i].getVarName (),
                helperVars [i].getTypeString (), outstream);
        }
        this.genAdditionalInstanceVars (fullName, baseName, outstream);
        outstream.println ();
    }
    /**
     *
     */
    protected void genInstanceVar (
        String varName,
        String typeString,
        PrintWriter outstream)
    {
        outstream.println ("    private " + typeString + " " + varName + ";");
    }
    /**
     *
     */
    void genConstructor (
        String fullName,
        String baseName,
        PrintWriter outstream)
    {
        outstream.println ("    public " + baseName + "(");
        outstream.print   ("        " + fullName + " _innerP");
        HelperVar [] helperVars = this.options.getHelperVars ();
        for (int i = 0; i < helperVars.length; ++i) {
            outstream.print (",\n        " + helperVars [i].getTypeString ()
                + " " + helperVars [i].getVarName () + "P");
        }
        outstream.println (")\n    {");
        outstream.println ("        this._inner = _innerP;");
        for (int i = 0; i < helperVars.length; ++i) {
            outstream.println ("        this." + helperVars [i].getVarName ()
                + " = " + helperVars [i].getVarName () + "P;");
        }
        outstream.println ("    }");
        outstream.println ();
    }
    /**
     *
     */
    String typeString (
        Class clazz)
    {
        String result = clazz.getName ();
        if (clazz.isArray ()) {
            char typeChar = result.charAt (1);
            switch (typeChar) {
                case 'Z':
                    result = "boolean []";
                    break;
                case 'B':
                    result = "byte []";
                    break;
                case 'C':
                    result = "char []";
                    break;
                case 'S':
                    result = "short []";
                    break;
                case 'I':
                    result = "int []";
                    break;
                case 'J':
                    result = "long []";
                    break;
                case 'F':
                    result = "float []";
                    break;
                case 'D':
                    result = "double []";
                    break;
                case 'L':
                    result = result.substring (2, result.length() - 1) + " []";
                    break;
                default:
                    result = result + " []";
                    break;
            }
        }
        return result;
    }
    /**
     *
     */
    boolean generatesClass (
        Class clazz)
    {
        boolean result = false;
        String clazzName = clazz.getName ();
        for (int i = 0; i < this.interfaceClasses.length; ++i) {
            String wrappedClass = this.interfacePackage + "." + this.interfaceClasses [i];
            if (clazzName.equals (wrappedClass)) {
                return true;
            }
        }
        return result;
    }
    /**
     *
     */
    protected void
    genImports (
        PrintWriter outstream)
    {
        outstream.println ("\n// nothing in genImports");
    }
    /**
     *
     */
    void genAdditionalInstanceVars (
        String fullName,
        String baseName,
        PrintWriter outstream)
    {
        outstream.println ("    // nothing in genAdditionalInstanceVars");
    }
    /**
     *
     */
    protected void genParameterEntry (
        Method method,
        PrintWriter outstream)
    {
        outstream.println ("        // nothing in genParameterEntry");
    }
    /**
     *
     */
    protected void genExceptionHandling (
        Method method,
        String exceptionVar,
        PrintWriter outstream)
    {
        outstream.println ("            // nothing in genExceptionHandling");
    }
    /**
     *
     */
    protected void genReturn (
        Method method,
        String resultVar,
        PrintWriter outstream)
    {
        outstream.println ("        // nothing in genReturn");
    }
    /**
     *
     */
    void genParameterPrint (
        Method method,
        String printerVar,
        PrintWriter outstream)
    {
        String delimiter = "";
        Class [] parameterTypes = method.getParameterTypes ();

        outstream.print ("        " + printerVar + ".println (this");
        outstream.print (" + \"." + method.getName () + " (\"");
        delimiter = " + arg";
        for (int i = 0; i < parameterTypes.length; ++i) {
            outstream.print (delimiter + i);
            delimiter = " + \", \" + arg";
        }
        outstream.println (" + \")\");");

    }
    /**
     *
     */
    void genExceptionPrint (
        Method method,
        String printerVar,
        String exceptionVar,
        PrintWriter outstream)
    {
        outstream.println ("            " + printerVar
            + ".println (" + exceptionVar + ".toString ());");
    }
    /**
     *
     */
    void genBacktracePrint (
        Method method,
        String printerVar,
        String exceptionVar,
        PrintWriter outstream)
    {
        outstream.println ("            " + exceptionVar
            + ".printStackTrace (" + printerVar + ");");
    }
    /**
     *
     */
    void genResultvarPrint (
        Method method,
        String printerVar,
        String resultVar,
        PrintWriter outstream)
    {
        outstream.println ("        " + printerVar
            + ".println (\"=> \" + " + resultVar + ");");
    }
}

